﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2019, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: BPlot2Dt (попиксельная визуализация зависящих от времени индикаторных функций, зашитых в шейдеры)
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, GLEW (http://glew.sourceforge.net/basic.html), OpenGL 2.1.
 * Клон BPlot2D.
 * Требует наличия x.bplot2dt.glsl.frag с текстом шейдера рядом с .exe (x -- цифра).
 *
 * Управление:
 * цифры: выбрать пиксельный шейдер (соответственно, файлы 0.bplot2dt.glsl.frag -- 9.bplot2dt.glsl.frag);
 * Стрелки и WASD: прокрутить экран;
 * +/=/Z, -/X: изменить масштаб;
 * HOME: вернуть экран в исходное положение;
 * ENTER и SPACE: перезагрузить текущий файл шейдера;
 * T: сбросить текущее время в нуль.
 * ESC: выйти;
 * Колёсико мыши изменяет масштаб (как +/-).
 */
#include <GL/glew.h>
#include <GL/freeglut.h>

#include <cstdint>
#include <cctype>
#include <ctime>

#include <utility>
#include <string>
#include <fstream>
#include <iomanip>
#include <algorithm>

#include <avo/Vec2.hpp>
#include <avo/Bbox2.hpp>
#include <avo/CoordinateGrid2.hpp>
#include <avo/TextDrawer.hpp>
#include <avo/Camera2.hpp>
#include <avo/Timer.hpp>
#include <avo/FileToString.hpp>


using Bbox_2 = avo::Camera_2::Box;
using Vec_2  = Bbox_2::Vec;


/// определение "сцены"
class Scene
{
  Bbox_2 bbox; // бокс, в который вписана сетка (для автопозиционирования камеры)

  avo::Camera_2 camera;  // камера
  avo::Coordinate_grid_2<double> cgrid; // координатная сетка
  mutable avo::Bitmap_text_drawer_2 text_drawer; // объект, выводящий текст
  int mouse_x, mouse_y; // последние зафиксированные координаты мыши

public:

  Scene() // бокс по умолчанию: 20x20
    : mouse_x(0), mouse_y(0), bbox(Vec_2(-10, -10), Vec_2(10, 10)) {}

  /// изобразить "сцену"
  void paint() const
  {
    cgrid.paint();
    draw_text();
  }

  /// получить бокс камеры
  const Bbox_2& camera_box() const
  {
    return camera.box();
  }

  /// приблизить камеру
  void camera_zoom_in()
  {
    camera.zoom_in(0.125);
    camera_changed();
  }

  /// "отодвинуть" камеру
  void camera_zoom_out()
  {
    camera.zoom_out(0.125);
    camera_changed();
  }

  /// передвинуть камеру
  void camera_move(const Vec_2 &delta)
  {
    camera.move(delta * camera.box().extents());
    camera_changed();
  }

  /// показать всю занятую область
  void camera_reset()
  {
    camera.reset(bbox);
    camera_changed();
  }

  /// изменён размер окна
  void reshape(int w, int h)
  {
    // изменить окно камеры
    camera.force_aspect((double)w / h);
    camera.apply();
    camera_changed();
    text_drawer.set_viewport(w, h);
  }

  /// установить координаты мыши
  void mouse_position(int x, int y)
  {
    mouse_x = x;
    mouse_y = y;
  }

private:

  /// обновить координатную сетку при изменении камеры
  void camera_changed()
  {
    cgrid.update(camera.box());
    text_drawer.set_coordinates(camera.box());
    camera.apply();
    glutPostRedisplay();  // уведомить FreeGlut, что следует перерисовать сцену
  }


  /// отображение координат и легенды графиков
  void draw_text() const
  {
    using std::to_string;
    auto smaller_box = camera;
    smaller_box.zoom_in(0.05); // чтобы не пропадали надписи на краях при округлении "наружу"
    auto &box = smaller_box.box();

    text_drawer.font(GLUT_BITMAP_HELVETICA_10);
    // размеры координатной сетки
    {
      // сгенерировать строки
      const auto left   = to_string(box.pmin.x),
                 bottom = to_string(box.pmin.y),
                 right  = to_string(box.pmax.x),
                 top    = to_string(box.pmax.y);

      // вывести текст средствами FreeGlut
      const auto center = 0.5 * (box.pmax + box.pmin);

      avo::LIGHT_GREY.select();
      text_drawer.view(box.pmin.x, center.y).put(left)
                 .view(center.x, box.pmin.y).put(bottom)
                 .view(box.pmax.x, center.y).put(right, avo::Alignment::toLeft)
                 .view(center.x, box.pmax.y).put(top, avo::Alignment::toDown)
                 ;
    }

    // координаты мыши
    {
      double mx, my; // "координаты" мыши
      text_drawer.pixels_to_view(mouse_x, mouse_y, mx, my);

      // сгенерировать строки
      const auto xy = to_string(mx) + '\n' + to_string(my);
    
      // вывести текст средствами FreeGlut
      avo::AZURE.select();
      text_drawer.view(box.pmin.x, box.pmax.y).put(xy, avo::Alignment::toDown);
    }
  }
};


//
// константы и глобальные переменные
//

/// имя файла пиксельного шейдера
/**
 * Подразумевает подмену первого символа.
 * Не слишком элегантное и масштабируемое решение, зато весьма простое :)
 * Предполагается, что шейдеров не может быть больше 10 (они привязаны к цифровым клавишам).
 */
char fragment_shader_filename[] = "0.bplot2dt.glsl.frag\0";

// индексы
GLuint program = -1,    ///< индекс программы (вершинный + пиксельный шейдеры), 
       vbuffer;         ///< индекс буфера вершин (прямоугольник на весь экран)

// индексы параметров шейдеров
GLint attribute_coord2d; ///< атрибут вершинного шейдера (координаты вершины на плоскости)
GLint uniform_screen;    ///< параметр для передачи размеров окна в пиксельный шейдер
GLint uniform_world;     ///< параметр для передачи системы координат (x0, y0, x1, y1)
GLint uniform_t;         ///< параметр для передачи текущего времени

// текущие значения параметров, передаваемые пиксельному шейдеру (время и размеры окна в пикселях)
float width = 800.f, height = 600.f;

/// момент начала рендеринга последнего кадра
double last_t, base_t;

/// поток вывода для сообщений об ошибках (файл)
std::ofstream err;

/// "сцена"
Scene scene;

/// количество вызовов repaint (для сообщений об ошибках -- указывается номер кадра)
long repaint_calls;


//
// вспомогательные функции
// осторожно: неявно используют err для сообщений об ошибках


/// уведомление об ошибках OpenGL
void log_gl_error(int point)
{
  // OpenGL 2.1 errors
  const char *msg = nullptr;
  const auto err_no = glGetError();
  switch (err_no)
  {
  case GL_NO_ERROR:
    return;
  case GL_INVALID_ENUM:
    msg = "invalid enum";
    break;
  case GL_INVALID_VALUE:
    msg = "invalid value";
    break;
  case GL_STACK_OVERFLOW:
    msg = "stack overflow";
    break;
  case GL_STACK_UNDERFLOW:
    msg = "stack underflow";
    break;
  case GL_OUT_OF_MEMORY:
    msg = "out of memory";
    break;
  case GL_TABLE_TOO_LARGE:
    msg = "table too large";
    break;
  default:
    msg = "unknown error";
  }

  err << "GL error #" << err_no << " (" << point
      << "), frame " << repaint_calls << ": " << msg << std::endl;
}


// получить индекс параметра шейдера, в качестве prog передаётся program
bool bind_uniform(GLuint prog, GLint &loc, const char *name)
{
  loc = glGetUniformLocation(prog, name);
  if (loc >= 0)
    return true;

  err << "Could not bind uniform \"" << name << "\".\n";
  return false;
}

// получить индекс атрибута шейдера
bool bind_attribute(GLuint prog, GLint &loc, const char *name)
{
  loc = glGetAttribLocation(program, name);
  if (loc >= 0)
    return true;

  err << "Could not bind attribute \"" << name << "\".\n";
  return false;
}


/// вывести сообщения об ошибках компиляции шейдера (компилятор входит в состав драйвера ГП) или компоновки программы
void log_errors(GLuint object)
{
  // получить длину текста сообщений
  GLint log_length = 0;
  if (glIsShader(object))
    glGetShaderiv(object, GL_INFO_LOG_LENGTH, &log_length);
  else if (glIsProgram(object))
    glGetProgramiv(object, GL_INFO_LOG_LENGTH, &log_length);
  else
    return (void)(err << "Checking shader: not a shader or a program.\n");

  if (log_length == 0)
    return;

  // подготовить место для текста
  std::string log(log_length, '\0');

  // получить текст сообщений
  if (glIsShader(object))
    glGetShaderInfoLog(object, log_length, NULL, &log[0]);
  else if (glIsProgram(object))
    glGetProgramInfoLog(object, log_length, NULL, &log[0]);

  // вывести сообщения в поток сообщений об ошибках
  err << "Shader errors:\n";
  err << log << '\n';
}


//
// управление ресурсами
//

/// отправить шейдерам значения параметров
void update_program_parameters()
{
  const auto &b = scene.camera_box();
  glUniform2f(uniform_screen, width, height);
  glUniform4f(uniform_world,
    (float)b.pmin.x,  (float)b.pmin.y,
    (float)b.width(), (float)b.height());
  glUniform1f(uniform_t, (float)last_t);
}


/// инициализация шейдеров и сборка программы (вершинный + фрагментный==пиксельный шейдеры)
bool init_program()
{
  if (program != -1)  // a no-op here actually
    glDeleteProgram(program);

  GLint compile_ok = GL_FALSE,
        link_ok    = GL_FALSE;
 
  // вершинный шейдер: тривиальный
  GLuint vs = glCreateShader(GL_VERTEX_SHADER);
  const char *vs_source = 
    "#version 120\n"  // OpenGL 2.1
    "attribute vec2 coord2d;                  "
    "void main(void)                          "
    "{                                        "
    "  gl_Position = vec4(coord2d, 0.0, 1.0); "
    "}                                        ";

  log_gl_error(1);
  glShaderSource(vs, 1, &vs_source, NULL);
  glCompileShader(vs);
  glGetShaderiv(vs, GL_COMPILE_STATUS, &compile_ok);
  if (!compile_ok)
  {
    err << "Can't compile the vertex shader.\n";
    log_errors(vs);
    return false;
  }

  log_gl_error(2);
  // пиксельный (фрагментный) шейдер: загружается из файла
  std::string fs_data = avo::file_to_string(fragment_shader_filename, err);
  if (fs_data.empty())
  {
    err << "File \"" << fragment_shader_filename << "\" (fragment shader source) not found or is empty.\n";
    return false;
  }

  err << "Loaded \"" << fragment_shader_filename << "\" successfully (" << fs_data.size() << " bytes).\n";
  const char *fs_source = fs_data.c_str(); // glShaderSource takes char* array

  GLuint fs = glCreateShader(GL_FRAGMENT_SHADER);
  glShaderSource(fs, 1, &fs_source, NULL);
  glCompileShader(fs);
  glGetShaderiv(fs, GL_COMPILE_STATUS, &compile_ok);
  if (!compile_ok)
  {
    err << "Can't compile the fragment shader.\n";
    log_errors(fs);
    return false;
  }

  log_gl_error(3);
  // компоновка программы (комбинация вершинного и пиксельного шейдеров)
  program = glCreateProgram();
  glAttachShader(program, vs);
  glAttachShader(program, fs);
  glLinkProgram(program);
  glGetProgramiv(program, GL_LINK_STATUS, &link_ok);
  if (!link_ok)
  {
    err << "Can't link shaders.\n";
    log_errors(program);
    return false;
  }

  // получить индексы атрибутов вершинного шейдера
  if (!bind_attribute(program, attribute_coord2d, "coord2d"))
    return false;

  // получить индексы параметров пиксельного шейдера
  if (!(bind_uniform(program, uniform_screen, "screen")
     && bind_uniform(program, uniform_world,  "world")
     && bind_uniform(program, uniform_t, "t")))
    return false;

  log_gl_error(4);
  return true;
}


/// инициализация используемых ресурсов (шейдеры, буферы, текстуры)
bool init_resources()
{
  if (!GLEW_VERSION_2_1)
  {
    err << "Sorry, we need at least OpenGL 2.1 support! You have "
        << glGetString(GL_VERSION) << ".\n";
    return false;
  } 

  // шейдеры
  if (!init_program())
    return false;

  // вершинный буфер

  // выделить новый индекс для буфера и включить этот буфер
  glGenBuffers(1, &vbuffer);
  glBindBuffer(GL_ARRAY_BUFFER, vbuffer);
  // будем рисовать прямоугольник на весь экран
  GLfloat vertices[] = 
  {
     -1.f,  -1.f,
      1.f,  -1.f,
     -1.f,   1.f,
      1.f,   1.f
  };

  log_gl_error(5);
  // передать данные и включить режим рисования из буфера, который больше не будет меняться (static)
  glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);
  glEnableVertexAttribArray(attribute_coord2d);

  glVertexAttribPointer(
    attribute_coord2d, // наш атрибут -- это пара координат x, y
    2,                 // количество элементов на вершины (две координаты)
    GL_FLOAT,          // тип элемента
    GL_FALSE,          // не включать автоматическую нормализацию
    0,                 // между группами данных нет пропусков
    (void*)(0)         // сдвиг начала данных в буфере равен нулю (используем буфер целиком)
  ); // приведение к указателю по историческим причинам (в OpenGL 1.x передавали указатель на массив в системной памяти)

  // считаем инициализацию успешной
  log_gl_error(6);
  return true;
}

/// удаление выделенных ресурсов
void free_resources()
{
  glDeleteBuffers(1, &vbuffer); // удалить вершинный буфер (прямоугольник)
  glDeleteProgram(program); // удалить шейдеры
}


//
// обработчики событий GLUT
//

/// перерисовка
void repaint()
{
  ++repaint_calls;
  // очистить экран
  glClear(GL_COLOR_BUFFER_BIT);
  // включить шейдер
  glUseProgram(program);
  log_gl_error(20);
  // передать текущие значения шейдеру
  last_t = avo::seconds() - base_t;
  update_program_parameters();
  log_gl_error(21);
  // нарисовать наш прямоугольник во весь экран
  glBindBuffer(GL_ARRAY_BUFFER, vbuffer);
  glEnableVertexAttribArray(attribute_coord2d);
  glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
  log_gl_error(22);
  // выключить шейдер
  glDisableVertexAttribArray(attribute_coord2d);
  glUseProgram(0);
  glBindBuffer(GL_ARRAY_BUFFER, 0);
  // нарисовать оси/текст
  glEnableClientState(GL_VERTEX_ARRAY);
  scene.paint();
  glDisableClientState(GL_VERTEX_ARRAY);
  log_gl_error(23);
  // показать результат на экране
  glFinish();
  glutSwapBuffers();
}


/// изменили размеры окна, нужно передать новые значения параметров в шейдер
void reshape(int w, int h)
{
  width   = (float)w;
  height  = (float)h;
  scene.reshape(w, h);
  glViewport(0, 0, w, h);
}


/// обработать нажатие "специальной" клавиши
void on_special_key(int key, int x, int y)
{
  Vec_2 delta; // 0, 0
  switch (key)
  {
  case GLUT_KEY_LEFT:   delta.x = -0.1; break;
  case GLUT_KEY_UP:     delta.y =  0.1; break;
  case GLUT_KEY_RIGHT:  delta.x =  0.1; break;
  case GLUT_KEY_DOWN:   delta.y = -0.1; break;
  case GLUT_KEY_HOME:   scene.camera_reset(); return;
  default:
    return; // ничего не делать, если была нажата какая-либо другая клавиша
  }

  scene.camera_move(delta);
}

/// обработка нажатий клавиш
void on_key(unsigned char key, int x, int y)
{
  switch(key) // код клавиши
  {
  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;

  case '+': case '=': case 'z': case 'Z':
    scene.camera_zoom_in(); break;
  case '-': case 'x': case 'X':
    scene.camera_zoom_out();  break;
  case 'w': case 'W':
    on_special_key(GLUT_KEY_UP, x, y); break;
  case 's': case 'S':
    on_special_key(GLUT_KEY_DOWN, x, y); break;
  case 'a': case 'A':
    on_special_key(GLUT_KEY_LEFT, x, y); break;
  case 'd': case 'D':
    on_special_key(GLUT_KEY_RIGHT, x, y); break;
  case 't': case 'T':
    base_t = avo::seconds(); break;
  case ' ': case '\n': case '\r': // перезагрузить текущий шейдер
    key = fragment_shader_filename[0];
    // дальше код под default делает то, что нужно

  default:
    if (std::isdigit(static_cast<unsigned char>(key)))
    {
      const char old_file = fragment_shader_filename[0];
      fragment_shader_filename[0] = key;
      // перезагрузить шейдер, пересобрать программу
      if (!init_program())
      {
        // если не удалось, вернуться к старому шейдеру
        fragment_shader_filename[0] = old_file;
        init_program();
      }
    }
  }
}


/// отслеживаем движение мыши, чтобы отображать её координаты
void on_hover(int x, int y)
{
  scene.mouse_position(x, (int)height - y);
  glutPostRedisplay();
}

/// масштабирование колёсиком мыши
void on_mouse_wheel(int wheel, int direction, int x, int y)
{
  if (direction < 0)
    scene.camera_zoom_in();
  else if (direction > 0)
    scene.camera_zoom_out();
}


/// вся анимация "спрятана" в функции перерисовки
void idle()
{
  avo::sleep(0.05);
  glutPostRedisplay();
}


//
// main
//
 
int main(int argc, char* argv[])
{
  // инициализация GLUT (создание окна)
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
  glutInitWindowSize(800, 600);
  glutCreateWindow("BPlot2Dt");
  err.open("BPlot2Dt.log");

  // вывести текущую дату и время
  {
    auto t = std::time(nullptr);
    err << "BPlot2Dt started " << std::put_time(std::localtime(&t), "%Y.%m.%d %H:%M:%S") << std::endl;
  }

  // инициализация GLEW (узнать возможности используемой реализации OpenGL)
  GLenum glew_status = glewInit();
  if (glew_status != GLEW_OK)
  {
    err << "GLEW error: " << (const char*)glewGetErrorString(glew_status) << ".\n";
    err.close();
    return EXIT_FAILURE;
  }

  // информация о контексте OpenGL и версии GLEW
  err << "GLEW version: "     << (const char*)glewGetString(GLEW_VERSION) << ".\n";
  err << "OpenGL implementation info\n\tVendor: "
                              << (const char*)glGetString(GL_VENDOR)
      << "\n\tRenderer: "     << (const char*)glGetString(GL_RENDERER)
      << "\n\tVersion: "      << (const char*)glGetString(GL_VERSION)
      << "\n\tGLSL version: " << (const char*)glGetString(GL_SHADING_LANGUAGE_VERSION) << "\n\n";
  
  // инициализация ресурсов, привязка обработчиков событий и старт цикла обработки сообщений
  if (init_resources())
  {
    glDisable(GL_DEPTH_TEST); // тест глубины здесь не нужен и будет только мешать
    glEnable(GL_BLEND);       // прозрачность
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    log_gl_error(10);
    
    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glClearColor(0.0f, 0.0f, 0.07f, 1.0f);
    log_gl_error(11);

    glutDisplayFunc(repaint);     // обработчик перерисовки
    glutIdleFunc(idle);           // постоянно вызывается в цикле

    glutReshapeFunc(reshape);     // обработчик изменения размеров окна
    glutKeyboardFunc(on_key); // обработчик нажатий клавиш
    glutSpecialFunc(on_special_key);
    glutPassiveMotionFunc(on_hover);
    glutMouseWheelFunc(on_mouse_wheel);

    scene.camera_reset();
    err.flush();
    glutMainLoop();               // запуск цикла обработки сообщений
    free_resources();             // освободить ресурсы
    err.close();
    return EXIT_SUCCESS;          // успешное завершение работы
  }

  // инициализация ресурсов не удалась
  err << "Initialization of GL resources failed.\n";
  err.close();
  return EXIT_FAILURE;
}
