/*
 * This file is part of AV Orchid.
 * Copyright (c) 2015, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file FIRModel.cpp
 * @author Кувшинов Д.Р.
 * @brief Реализация FIRModel.hpp. Собственно FIR-фильтр, вычисляющий производную, находится здесь.
 */
#include "FIRModel.hpp"
#include <cstdlib>
#include <cstring>
#include <cassert>
#include <intrin.h> // MSVC: для поддержки SSE3

using namespace std;

/// FIR-фильтр: свёртка пяти значений, производная, O(h4), сдвиг на два отсчёта.
/// "Универсальный вариант".
template <class Float, class InIt, class OutIt>
OutIt derivative_4(Float step, InIt from, InIt to, OutIt dest)
{
  const Float
    bx = Float(1) / (12 * step),
    by = Float(2) / (3 * step);

  if (from == to) return dest;
  auto f0 = *from;
  if (++from == to) return dest;
  auto f1 = *from;
  if (++from == to) return dest;
  auto f2 = *from;
  if (++from == to) return dest;
  auto f3 = *from;
  if (++from == to) return dest;
  auto f4 = *from;
  
  while (true)
  {
    *dest = bx * (f0 - f4) - by * (f1 - f3);
    ++dest;

    if (++from == to)
      return dest;

    f0 = f1;
    f1 = f2;
    f2 = f3;
    f3 = f4;
    f4 = *from;
  }
}


/// FIR-фильтр: свёртка пяти значений, производная, O(h4), сдвиг на два отсчёта, SSE3-реализация.
float* derivative_4sse3(float step, const float *from, const float *to, float *dest)
{
  if (to - from < 5)
    return dest;

  auto
    f0123 = _mm_loadu_ps(from),
    f5 = _mm_load_ss(from + 4);
  from += 5;

  const float
    bx = 1.f / (12.f * step),
    by = 2.f / (3.f * step);

  const auto
    coefs = _mm_set_ps(bx, -by, by, -bx);

  const auto
    f2_34 = _mm_unpacklo_ps(f5, f0123);

  auto
    f2222 = _mm_shuffle_ps(f0123, f0123, _MM_SHUFFLE(1, 1, 1, 1)), // f2 copied
    f0134 = _mm_shuffle_ps(f2_34, f0123, _MM_SHUFFLE(3, 2, 1, 0));
  
  while (true)
  {
    // свёртка
    const auto
      mult = _mm_mul_ps(f0134, coefs),
      sum2 = _mm_hadd_ps(mult, mult),  // SSE3
      sum1 = _mm_add_ss(sum2, _mm_shuffle_ps(sum2, sum2, _MM_SHUFFLE(1, 1, 1, 1)));

    _mm_store_ss(dest++, sum1);
    if (from == to)
      return dest;

    // считывание следующего элемента и сдвиг регистра
    f5 = _mm_load_ss(from++);

    const auto
      f__45 = _mm_unpacklo_ps(f5, f0134),
      f12__ = _mm_unpackhi_ps(f2222, f0134);
    
    f2222 = _mm_shuffle_ps(f0134, f0134, _MM_SHUFFLE(1, 1, 1, 1)); // старый третий -- новый второй
    f0134 = _mm_shuffle_ps(f__45, f12__, _MM_SHUFFLE(3, 2, 1, 0));
  }
}


/// Растущий буфер с областями записи и чтения.
/// Не thread-safe.
/// Использует функции C для управления памятью (calloc, realloc, free).
class Growing_buffer
{
  float* sp; // starting point, начало хранилища
  float* rp; // reading point, текущая позиция чтения
  float* wp; // writing point, текущая позиция записи, invariant: wp >= rp >= sp
  size_t sz; // полный размер хранилища

  void nullify()
  {
    sp = nullptr;
    rp = nullptr;
    wp = nullptr;
    sz = 0;
  }

  Growing_buffer(Growing_buffer&) = delete;
  Growing_buffer& operator=(Growing_buffer&) = delete;

public:

  /// Создаёт пустой буфер без хранилища.
  Growing_buffer()
    : sp(nullptr), rp(nullptr), wp(nullptr), sz(0) {}

  Growing_buffer(Growing_buffer &&other)
    : sp(other.sp), rp(other.rp), wp(other.wp), sz(other.sz)
  {
    other.nullify();
  }

  /// Пытается выделить буфер на size чисел.
  /// Новый буфер заполняет нулями.
  explicit Growing_buffer(size_t size)
    : Growing_buffer()
  {
    if (wp = rp = sp = static_cast<float*>(calloc(size, sizeof(float))))
      sz = size;
  }

  ~Growing_buffer()
  {
    free(sp);
  }

  Growing_buffer& operator=(Growing_buffer &&other)
  {
    if (this != &other)
    {
      free(sp);
      sp = other.sp;
      rp = other.rp;
      wp = other.wp;
      sz = other.sz;
      other.nullify();
    }
    return *this;
  }

  /// Узнать размер области чтения (сколько элементов можно прочитать).
  size_t available_elements() const { return wp - rp; }

  const float* read() const { return rp; }
  
  /// Возвращает позицию чтения на момент вызова.
  /// Передвигает позицию чтения на take_items вперёд.
  /// Если take_items слишком велико, возвращает nullptr, не изменяя позицию чтения.
  const float* read(size_t take_items)
  {
    const auto old_rp = rp;
    if (take_items > static_cast<size_t>(wp - old_rp))
      return nullptr;
    
    rp = old_rp + take_items;
    return old_rp;
  }

  /// Возвращает позицию начала записи с гарантированным объёмом в new_items.
  /// В случае невозможности выделить достаточное количество памяти возвращает nullptr.
  /// Значение guard задаёт количество дополнительных элементов до текущей позиции чтения,
  /// которые необходимо сохранить в буфере.
  float* write(size_t new_items, size_t guard = 0)
  {
    assert(rp - sp >= (ptrdiff_t)guard);
    const auto old_wp = wp;
    // достаточно места в конце буфера для добавления new_items?
    if (new_items <= sz - (old_wp - sp))
    {
      wp = old_wp + new_items;
      return old_wp;
    }

    // сдвигаем хвост в начало, убирая уже прочитанные элементы из буфера
    const auto rp_rest = guard + (old_wp - rp);
    memmove(sp, rp - guard, rp_rest);
    
    // достаточно ли сдвига для добавления new_items?
    if (new_items <= sz - rp_rest)
    {
      rp = sp + guard;
      const auto old_wp = sp + rp_rest;
      wp = old_wp + new_items;
      return old_wp;
    }

    // если не достаточно -- realloc
    {
      const auto new_sz = new_items + rp_rest;
      const auto new_sp = static_cast<float*>(realloc(sp, new_sz));
      if (!new_sp)
        return nullptr; // failed to allocate

      sz = new_sz;
      sp = new_sp;
      rp = new_sp + guard;
      const auto old_wp = new_sp + rp_rest;
      wp = old_wp + new_items;
      return old_wp;
    }
  }
};


// pimpl, состояние модели
struct FIR_model::FIR_model_state
{
  FIR_model_state(FIR_model_state&) = delete;
  FIR_model_state& operator=(FIR_model_state&) = delete;

  static const unsigned BUFFERS = 2;
  Growing_buffer buffer[BUFFERS];
  Signal_generator signal_gen;

  float last_t;
  const float timestep;
  
  FIR_model_state(size_t bufsz, float timestep)
    : signal_gen([](float) { return 0.f; }), last_t(0.f), timestep(timestep)
  {
    for (auto &buf: buffer)
      buf = Growing_buffer(bufsz);
  }
};


void FIR_model::cleanup()
{
  delete state;
  state = nullptr;
}


FIR_model::FIR_model(size_t bufsz, float timestep)
  : state(nullptr)
{
  assert(bufsz != 0);
  state = new FIR_model_state(bufsz, timestep);
}


void FIR_model::select_signal(FIR_model::Signal_generator sg)
{
  assert(state);
  state->signal_gen = sg;
}


size_t FIR_model::next_time(float t)
{
  assert(state);
  const auto
    t0 = state->last_t,
    th = state->timestep;

  if (t <= t0)
    return 0;

  state->last_t = t;
  const auto samples = 
    static_cast<size_t>(
    static_cast<ptrdiff_t>(
      (t - t0) / th
    ));

  assert(samples < (1 << 30));
  // generate signal
  const auto swp = state->buffer[0].write(samples, 4);
  const auto sg = state->signal_gen;
  for (size_t i = 0; i < samples; ++i)
    swp[i] = sg(th * i);

  // generate derivatives
  const auto dwp = state->buffer[1].write(samples);
  derivative_4(th, swp - 4, swp + samples, dwp);
  
  return samples;
}


std::size_t FIR_model::available_samples(Buffer buffer) const
{
  const auto buf = static_cast<unsigned>(buffer);
  if (buf >= FIR_model_state::BUFFERS || !state)
    return 0;
  return state->buffer[buf].available_elements();
}


const float* FIR_model::take_samples(FIR_model::Buffer buffer, size_t samples)
{
  const auto buf = static_cast<unsigned>(buffer);
  if (buf >= FIR_model_state::BUFFERS)
    return nullptr;
  assert(state);
  return state->buffer[buf].read(samples);
}
