﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file  Mesh2.hpp
 * @brief полигональный примитив для OpenGL
 */
#pragma once
#include <avo/Color.hpp>
#include <avo/Prim2.hpp>
#include <type_traits>
#include <vector>

#include "Contour.hpp"

using Prim_2 = avo::Prim_2<Scalar>;

/// набор примитивов, задающий произвольный плоский многоугольник
class Mesh_2
{
  std::vector<Prim_2>
    excont, ///< внешние контуры (каждый ограничивает компоненту связности)
    incont, ///< внутренние контуры ("дырки")
    body    ///< треугольники, на которые разбита внутренность многоугольника (триангуляция)
  ;

  /// выставляет цвета по умолчанию
  void setDefaultColors()
  {
    externalOutline = avo::YELLOW;
    internalOutline = avo::RED;
    background = avo::GREEN;
    background.alpha = 0.4f;
  }

public:

  avo::Color
    externalOutline, ///< цвет внешних контуров
    internalOutline, ///< цвет внутренних контуров
    background       ///< цвет заливки внутренности многоугольника
  ;


  /// конструктор по умолчанию
  Mesh_2()
    : tess(nullptr)
  {
    setDefaultColors ();
  }

  /// построить многоугольник на основе набора контуров
  /**
   * Выполняет триангуляцию и сортировку контуров (на внешние и внутренние).
   */
  template <class ContourForIt>
  Mesh_2(ContourForIt contbegin, ContourForIt contend)
    : tess(nullptr)
  {
    using std::begin;
    using std::end;

    setDefaultColors();
    if (contbegin != contend)
    {
      start_polygon();
      // добавить каждый контур в триангуляцию
      for (; contbegin != contend; ++contbegin)
      {
        auto pbegin = begin(*contbegin),
             pend   = end(*contbegin);

        if (pbegin != pend) // контур не пуст
        {
          // определить ориентацию контура по знаку площади
          ( contbegin->area() < 0.0? incont: excont )
            .emplace_back(pbegin, pend, GL_LINE_LOOP);

          // добавить контур в разбиение
          start_contour();
          for (; pbegin != pend; ++pbegin) // передать все вершины контура
            push_point(*pbegin);
          end_contour();
        }
      }
      end_polygon();
    }
  }

  // compiler might have defined the following ctors and op=s automatically if a destructor wasn't there

  Mesh_2(const Mesh_2 &other)
    : excont(other.excont),
      incont(other.incont),
      body(other.body),
      externalOutline(other.externalOutline),
      internalOutline(other.internalOutline),
      background(other.background),
      tess(nullptr)
  {}

  Mesh_2(Mesh_2 &&other)
    : excont(std::move(other.excont)),
      incont(std::move(other.incont)),
      body(std::move(other.body)),
      externalOutline(other.externalOutline),
      internalOutline(other.internalOutline),
      background(other.background),
      tess(nullptr)
  {}

  Mesh_2& operator=(const Mesh_2 &other)
  {
    excont = other.excont;
    incont = other.incont;
    body = other.body;
    externalOutline = other.externalOutline;
    internalOutline = other.internalOutline;
    background = other.background;
    return *this;
  }

  Mesh_2& operator=(Mesh_2 &&other)
  {
    excont = std::move(other.excont);
    incont = std::move(other.incont);
    body = std::move(other.body);
    externalOutline = other.externalOutline;
    internalOutline = other.internalOutline;
    background = other.background;
    return *this;
  }


  // наличие явно определённого деструктора -- банальная перестраховка
  ~Mesh_2()
  {
    release_tessellator(); // избыточно...
  }


  /// обмен содержимым
  void swap(Mesh_2 &other);

  /// отрисовка средствами OpenGL
  void paint() const
  {
    block_paint (body.begin(), body.end(), background);
    block_paint (incont.begin(), incont.end(), internalOutline);
    block_paint (excont.begin(), excont.end(), externalOutline);
  }

private:

  /// нарисовать набор примитивов общим цветом
  template <class ForIt>
  static void block_paint(ForIt begin, ForIt end, const avo::Color &color)
  {
    color.select();
    for (; begin != end; ++begin)
      begin->paint();
  }

  // используется только в конструкторе из контуров, в остальное время == nullptr
  class Tessellator *tess;

  void release_tessellator();
  void start_polygon(); ///< выполняет также привязку колбеков GLU
  void end_polygon();   ///< выполняет отвязку колбеков GLU
  void start_contour();
  void end_contour();
  void push_point(const Vec_2&);
};


inline void swap(Mesh_2 &a, Mesh_2 &b)
{
  a.swap(b);
}
