/*
 * This file is part of AV Orchid.
 * Copyright (c) 2014, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Basics.hpp
 * @author Dmitri Kuvshinov
 * @brief different minor general stuff
 */
#pragma once
#include <cstddef>
#include <cstdlib>
#include <cassert>

namespace avo
{

/// next pseudo-random float from [0, 1]
inline float random()
{
  static const float inverse_rand_max = 1.0f / RAND_MAX;
  return std::rand() * inverse_rand_max;
}

/// next pseudo-random something from [a, b]
template <class T>
inline T random(const T a, const T b)
{
  static const double inverse_rand_max = 1.0 / RAND_MAX;
  return a + static_cast<T>(double(b - a) * (std::rand() * inverse_rand_max));
}


/// clamp a value to the range given
template <class T>
inline T clamp(T val, T min_val = T(0), T max_val = T(1))
{
  if (val < min_val)
    return min_val;
  if (max_val < val)
    return max_val;
  return val;
}

/// linear interpolation, if t is 0 then return a, if t is 1 then return b
template <class T, class X>
inline T lerp(const T &a, const T &b, X t)
{
  return (X(1) - t) * a + t * b;
}

/// treat an array as a piecewise linear function with uniform nodes at 0...size-1
template <class A, class X>
inline auto piecewise_sample_array(A array, std::size_t array_size, X t)
    -> decltype(lerp(array[0], array[0], t))
{
  assert(array_size > 0);
  const auto lower_t = static_cast<std::ptrdiff_t>(t),
             upper_t = lower_t + 1;

  if (lower_t < 0)
    return array[0]; // first
  if (array_size <= static_cast<std::size_t>(upper_t)) // it's not negative here
    return array[array_size - 1]; // last
  return lerp(array[lower_t], array[upper_t], t - static_cast<X>(lower_t));
}

}
