/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Color.hpp
 * @author Dmitri Kuvshinov
 * @brief RGBA color over float
 */
#pragma once
#include <GL/freeglut.h>
#include "Basics.hpp"


namespace avo
{

/// RGBA color with fp32 components
struct Color
{
  float red   ///< red channel value
      , green ///< green channel value
      , blue  ///< blue channel value
      , alpha ///< alpha channel value (usually used as opaqueness value, i.e. 0.0 means transparent, 1.0 means opaque)
      ;

  /// construct a shade of gray (0.0 for black, 1.0 for white)
  explicit Color(float gray = 1.0f, float a = 1.0f)
    : red(gray), green(gray), blue(gray), alpha(a) {}

  /// construct a color value from channel values
  Color(float r, float g, float b, float a = 1.0f)
    : red(r), green(g), blue(b), alpha(a) {}

  /// construct a color value by combining RGB part of another color and a given alpha value
  Color(const Color &rgb, float alpha)
    : red(rgb.red), green(rgb.green), blue(rgb.blue), alpha(alpha) {}

  /// select this color in the OpenGL compatibility context
  void select() const
  {
    glColor4fv(&red);
  }

  /// select this color as clear color
  void select_clear() const
  {
    glClearColor(red, green, blue, alpha);
  }

  /// make a shade of gray by computing luminance (Y component of ITU-R recommendation)
  float luminance() const
  {
    return 0.299f * red + 0.587f * green + 0.114f * blue;
  }

  /// clamp all components to [0, 1]
  Color clamped() const
  {
    return Color(clamp(red), clamp(green), clamp(blue), clamp(alpha));
  }

  /// make color bright by normalizing its maximal component (while "black" becomes white)
  Color bright() const
  {
    // I think this code is funny
    const float one = 1.0f;
    const float r = red, g = green, b = blue;

    const auto b_le_g = (b <= g? 1: 0),
               g_le_r = (g <= r? 1: 0),
               r_le_g = (r <= g? 1: 0),
               b_le_r = (b <= r? 1: 0);

    if (b_le_r & g_le_r)
      goto red_max;
    if (b_le_g & r_le_g)
      goto green_max;    

    //blue_max:
    {
      const float rb = one / b;
      return Color(r * rb, g * rb, one);
    }

    red_max:
    {
      if (r != 0.0)
      {
        const float rr = one / r;
        return Color(one, g * rr, b * rr);
      }
      return Color(one); // and black becomes white!
    }

    green_max:
    {
      const float rg = one / g;
      return Color(r * rg, one, b * rg);
    }
  }

  /// color vector addition (alpha may have some problems)
  Color& operator+=(const Color &other)
  {
    red   += other.red;
    green += other.green;
    blue  += other.blue;
    alpha += other.alpha;
    return *this;
  }

  /// color vector subtraction (alpha may have some problems)
  Color& operator-=(const Color &other)
  {
    red   -= other.red;
    green -= other.green;
    blue  -= other.blue;
    alpha -= other.alpha;
    return *this;
  }

  /// modulate color
  Color& operator*=(float coef)
  {
    red   *= coef;
    green *= coef;
    blue  *= coef;
    alpha *= coef;
    return *this;
  }
};


inline bool operator==(const Color &a, const Color &b)
{
  return a.red == b.red  && a.green == b.green
     && a.blue == b.blue && a.alpha == b.alpha;
}

inline bool operator!=(const Color &a, const Color &b)
{
  return !(a == b);
}

inline Color operator+(const Color &a, const Color &b)
{
  return Color(a) += b;
}

inline Color operator-(const Color &a, const Color &b)
{
  return Color(a) -= b;
}

inline Color operator*(float coef, const Color &b)
{
  return Color(b) *= coef;
}

inline Color operator*(const Color &b, float coef)
{
  return coef * b;
}


// predefined colors
const Color
    RED         (1.0f, 0.0f, 0.0f)
  , ORANGE      (1.0f, 0.6f, 0.0f)
  , YELLOW      (1.0f, 1.0f, 0.0f)
  , LIME        (0.6f, 1.0f, 0.0f)
  , GREEN       (0.0f, 1.0f, 0.0f)
  , TEAL        (0.0f, 1.0f, 0.6f)
  , CYAN        (0.0f, 1.0f, 1.0f)
  , AZURE       (0.4f, 0.6f, 1.0f)
  , BLUE        (0.0f, 0.0f, 1.0f)
  , VIOLET      (0.4f, 0.0f, 0.8f)
  , MAGENTA     (1.0f, 0.0f, 1.0f)
  , PINK        (1.0f, 0.3f, 0.5f)
  , SALMON      (1.0f, 0.6f, 0.5f)
  , OLIVE       (0.5f, 0.4f, 0.2f)
  , CHOCO       (0.2f, 0.1f, 0.0f)
  , BLACK       (0.0f, 0.0f, 0.0f)
  , DARK_GREY   (0.3f, 0.3f, 0.3f)
  , GREY        (0.5f, 0.5f, 0.5f)
  , LIGHT_GREY  (0.8f, 0.8f, 0.8f)
  , WHITE       (1.0f, 1.0f, 1.0f)
  ;

}
