﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Pixels
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Демонстрация вывода произвольной пиксельной графики средствами OpenGL.
 * Пиксель задаётся 4 байтами ("32-битный цвет"), каждый байт отвечает одному каналу цвета:
 * r -- red == "красный", g -- green ==  "зелёный", b -- blue == "синий" и a -- "альфа"
 * (обычно задаёт прозрачность, здесь не используется).
 *
 * Картинка представляет собой прямоугольный массив (непрерывный длины "высота" * "ширина") пикселей.
 * В данном примере вывод картинки в окно выводится вызовом команды OpenGL glDrawPixels.
 *
 * На старте программы создаётся картинка фиксированного размера.
 * Нажатие ENTER/SPACE присваивает пикселям псевдослучайные значения.
 * Нажатие ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>
#include <cstdlib>  // rand/srand
#include <ctime>    // time
#include <iostream>

#include <avo/Pixel.hpp> // RGBA-цвет в 32-битном формате (компоненты типа uint8_t)

using avo::Pixel;

//
// константы и глобальные переменные
//

/// размеры картинки: ширина и высота (для простоты зафиксированы)
const int width = 800, height = 600;
/// сама картинка
Pixel pixels[height][width];


//
// вспомогательные функции
//

/// следующее псевдослучайное целое [minv, maxv]
inline Pixel random_pixel()
{
  return Pixel::rgb(std::rand() % 256, std::rand() % 256, std::rand() % 256);
}

/// перезаписать пиксели случайными значениями
void new_pixels()
{
  #pragma omp parallel for // каждый пиксель обрабатывается независимо, можно распараллелить по строчкам
  for (int i = 0; i < height; ++i)
    for (int j = 0; j < width; ++j)
      pixels[i][j] = random_pixel();
      // На самом деле, здесь скрыта потенциальная ошибка:
      // rand изменяет скрытое глобальное состояние генератора псевдослучайных чисел,
      // одновременные вызовы из нескольких потоков могут вызвать повреждение этого состояния.
      // Впрочем, текущая многопоточная версия MSVC CRT должна корректно обрабатывать эту ситуацию.

  // TASK: вместо "шума" нарисовать "шахматную доску".
}


//
// обработчики событий GLUT
//

/// перерисовка
void repaint()
{        /* ширина,  высота, тип пикселя, тип компонента, указатель на картинку */
  glDrawPixels(width, height, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
  glFinish(); // дождаться завершения обработки очереди команд OpenGL
  // обмен буферов не нужен, так как здесь буфер всего один
}


/// вычислить следующий кадр
void frame_move()
{
  #pragma omp parallel for // можно обрабатывать строки параллельно, так как они не зависят друг от друга,
  for (int i = 0; i < height; ++i) // в процессе обработки одной строки меняется состояние лишь самой этой строки
  {
    Pixel *line   = pixels[i]; // текущая строчка как массив
    // одномерное "окно": два предыдущих пикселя
    Pixel left    = line[0],   // пиксель слева
          middle  = line[1];   // текущий пиксель

    for (int j = 2; j < width; ++j) // выполнить простой фильтр (горизонтальное усреднение с радиусом 1 пиксель)
    {
      const Pixel right = line[j]; // пиксель справа от текущего

      // записать новый пиксель вместо текущего среднего
      line[j - 1] = Pixel::rgb
      ( // благодаря умножению на 2 автоматически приводится к int, что помогает избежать раннего переполнения
        (2 * middle.r + left.r + right.r) / 4, // красный
        (2 * middle.g + left.g + right.g) / 4, // зелёный
        (2 * middle.b + left.b + right.b) / 4  // синий
      );
      
      // сдвинуть окно фильтра вперёд на один пиксель
      left   = middle;
      middle = right;
    }
  }
  // TASK: реализовать двумерный фильтр сглаживания, например, такой:
  //  1/4  1  1/4
  //  1    2  1    *1/7
  //  1/4  1  1/4
  // Подсказка: записывать результат в отдельный массив, а не выполнять обработку "на месте".

  glutPostRedisplay(); // попросить перерисовку на следующей итерации главного цикла
}


/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void on_key(unsigned char key, int x, int y)
{
  switch(key) // код клавиши
  {
  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;

  case ' ': case '\r': case '\n':
    new_pixels(); // заполнить псевдослучайными цветами
    break;
  
  default:
    return;
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv);  // обработка параметров командной строки
  std::srand((unsigned)std::time(nullptr)); // рандомизация с помощью текущего системного времени

  // создать окно с привязанным контекстом OpenGL
  glutInitDisplayMode(GLUT_RGBA | GLUT_SINGLE); // режим: 32-битный цвет RGBA, одинарный буфер кадра SINGLE
  glutInitWindowPosition(0, 0);      // координаты верхнего левого угла окна на экране
  glutInitWindowSize(width, height); // полные ширина и высота окна
  glutCreateWindow("Pixels");        // создать окно, в функцию передаётся заголовок создаваемого окна

  // привязка функций-обработчиков событий
  glutIdleFunc(frame_move); // вызывать, когда нет событий
  glutDisplayFunc(repaint); // вызывать для отрисовки кадра
  glutKeyboardFunc(on_key); // вызывать для обработки нажатий "обычных" клавиш
  
  // старт
  std::cout << "Starting Pixels\n";
  new_pixels();
  glutMainLoop(); // войти в цикл обработки сообщений ("главный цикл")
  // после выхода из цикла обработки сообщений исполнение программы продолжается отсюда
  std::cout << "\nbye!";
  return 0;
}
