/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file   Timer.cpp
 * @brief  Timer.hpp implementation
 * @author Dmitri Kuvshinov
 */
#include <avo/Timer.hpp>

//
// MSVC 2013 and lower don't support chrono high resolution timer,
// we'll use the one provided in Win32
#if defined(WIN32) && defined(_MSC_VER) && _MSC_VER <= 1800

#define WIN32_LEAN_AND_MEAN // NO*** stuff might have been defined too
#include <Windows.h>


namespace
{
  template<class To, class From>
  inline To* pointer_cast(From *p) // actually, I can't say if it's in any way better than
  { return static_cast<To*>(static_cast<void*>(p)); } // just doing basic reinterpret_cast

  struct HPT
  {
    long long start;
    double freq;

    HPT()
      : freq(1.0), start(1)
    {
      long long fr = 1;
      QueryPerformanceFrequency (pointer_cast<LARGE_INTEGER>(&fr));
      freq = static_cast<double>(fr);
      QueryPerformanceCounter (pointer_cast<LARGE_INTEGER>(&start));
    }

    double operator()() const
    {
      long long t = 0;
      QueryPerformanceCounter (pointer_cast<LARGE_INTEGER>(&t));
      return static_cast<double>(t - start) / freq;
    }
  } Hpt; // high-performance timer, global static (hidden) object
  // should be thread-safe, it's state is not being changed after creation
}

double avo::seconds()
{
  return Hpt();
}

double avo::timer_ticks()
{
  return Hpt.freq;
}

void avo::sleep(double t)
{
  Sleep(static_cast<DWORD>(1000.0 * t));
}


#else
// standard-way

#include <chrono>
#include <thread>

namespace
{
  const auto starting_time = std::chrono::high_resolution_clock::now();
}

double avo::seconds()
{
  using namespace std::chrono;
  return (duration_cast<nanoseconds>(high_resolution_clock::now() - starting_time)).count() * 1.0e-9;
}

double avo::timer_ticks()
{
  return 1.0e+9; // pseudoticks, nanoseconds
}

void avo::sleep(double t)
{
  std::this_thread::sleep_for(std::chrono::milliseconds(static_cast<int>(1000.0 * t)));
}

#endif
