﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file ConfigFile.cpp
 *
 * Средства для чтения файлов-словарей, состоящих из строчек вида
 * "ключ" = "значение" комментарий
 *
 * Кавычки не обязательны, но если их нет, то читается только одно "слово".
 * Например, строка
 * > my name = Zorg
 * игнорируется.
 * Предполагается, что ключи и значения -- строки.
 */
#include "ConfigFile.hpp"
#include <sstream>
#include <fstream>
#include <algorithm>
#include <iterator>
#include <cstring>

using namespace std;


namespace
{

// запись строки в виде литерала

/// пишет 8-ричный код ch, например, writeOct(cout, ' ') должно вывести \040
inline void writeOct(ostream &os, unsigned char c)
{
  os << '\\' << ('0' + c / 64) << ('0' + (c / 8) % 8) << ('0' + c % 8);
}

/// пишет представление символа ch в поток os
void write(ostream &os, unsigned char ch)
{
  if (ch < 32 || ch == 255)
  {
    writeOct(os, ch);
  }
  else
  {
    switch (ch)
    {
    case '\\': os << "\\\\"; break; // need moar slashes!!
    case '"' : os << "\\\""; break;
    default:   os << ch;
    }
  }
}

/// записать строку в виде C-литерала
void write_literal(ostream &os, const string &s)
{
  os << '"';
  for (auto c : s)
    write(os, static_cast<unsigned char>(c));
  os << '"';
}

/// проверить, содержит ли строка пробельные символы
inline bool contains_spaces(const string &s)
{
  return s.find_first_of(" \t\n\r\f\v") != string::npos; 
}

/// записать строку в виде слова или C-литерала
inline void write(ostream &os, const string &s)
{
  if (contains_spaces(s))
    write_literal(os, s);
  else
    os << s;
}


//
// разбор литерала как строки, ASCII

/// проверить, является ли ch 8-ричной цифрой
inline bool is_oct_digit(char ch)
{
  return (unsigned)(ch - '0') < 8u;
}

/// получить числовое значение ch как 8-ричной цифры
inline int oct_to_int(char ch)
{
  return ch - '0';
}

/// проверить, является ли ch 16-ричной цифрой
inline bool is_hex_digit(char ch)
{
  return strchr("0123456789abcdefABCDEF", ch) != nullptr;
}

/// получить числовое значение ch как 16-ричной цифры
inline int hex_to_int(char ch)
{
  if (ch <= '9') return ch - '0';
  return ch - ('a' <= ch && ch <= 'f'? ('a' - '\x0A') : ('A' - '\x0A'));
}


/// чтение C-литерала из потока
string parse_literal(istream &is)
{
  char ch = 0;
  is >> ch; // пропускает ведущие пробелы
  if (!is) return "";

  string result; // прочитанная строка
  if (ch != '"') // одиночное "слово"
  {
    is.putback(ch);
    is >> result;
  }
  else // литерал, ad-hoc parser, beware! it hasn't been tested well
  {
    for (;;) // читает до закрывающей " или конца потока
    {
      ch = is.get();
      if (!is || ch == '"') break;
      
      if (ch == '\\') // пришла \ -- прочитать спецсимвол
      {
        switch (ch = is.get())
        {
        case 'n':  ch = '\n'; break;
        case 'r':  ch = '\r'; break;
        case '\\': ch = '\\'; break;
        case '"':  ch = '"';  break;
        case '\'': ch = '\''; break;
        case 't':  ch = '\t'; break;
        case 'b':  ch = '\b'; break;
        case 'a':  ch = '\a'; break;
        case 'f':  ch = '\f'; break;
        case 'v':  ch = '\v'; break;
        case '?':  ch = '?';  break;
        case 'x': // hex-code
          {
            int code = 0;
            while (is_hex_digit(ch = is.get()))
              code = code * 16 + hex_to_int(ch);
            if (is)
              is.putback(ch);
            ch = code; // теряет многобайтный символ
          }
          break;

        default: // octal-code? до трёх цифр
          if (is_oct_digit(ch))
          {
            int code = oct_to_int(ch);
            for (int i = 0; i < 3 && is_oct_digit(ch = is.get()); ++i)
              code = code * 8 + oct_to_int(ch);
            if (is && !is_oct_digit(ch))
              is.putback(ch);
            ch = code;
          }
          else // не восьмиричный код, просто кладём прочитанные символы в результат
          {
            result += '\\';
            // ch будет добавлен следующим, 
            // однако, возможно, мы уже дошли до конца потока, тогда ch == EOF и можно выйти
          }
        }
      }

      if (!is) break;
      result += ch;
    }      
  }

  return result;
}


/// результат попытки чтения элемента словаря
enum Read_entry_result { EntryReadSuccessful, EntryReadEOF, EntryReadSkip };
/// чтение элемента словаря (пара ключ, значение)
/**
 * @param tail после завершения содержит неразобранный остаток строки (либо всю строку, если разобрать не удалось)
 */
Read_entry_result read_entry(istream &is, string &key, string &value, string &tail)
{
  string line;
  getline(is, line);
  if (!is) return EntryReadEOF;
  // данный подход не является наиболее эффективным по скорости/памяти,
  // зато упрощает программирование
  istringstream ls(line);

  // ключ
  key = parse_literal(ls);
  if (key.size() == 0 || key.size() > 1 && key.substr(0, 2) == "//") // пропустить
    return tail = move(line), EntryReadSkip;

  // =
  char equals = 0;
  ls >> equals;
  if (equals != '=') // пропустить
    return tail = move(line), EntryReadSkip;

  // значение
  value = parse_literal(ls);

  // хвост
  getline(ls, tail);
  return EntryReadSuccessful;
}


/// считать словарь из потока
Dictionary read_dictionary(istream &is)
{
  Dictionary result;
  string key, value, tail;
  for (;;)
  {
    switch (read_entry(is, key, value, tail))
    {
    case EntryReadSuccessful:
      result.emplace(move(key), move(value));
      break;
    case EntryReadEOF:
      return result;
    default: {} // ignore EntryReadSkip
    }
  }
}


/// считать словарь из одного потока, заменить значения и записать в другой поток
void refresh_dictionary(istream &input, ostream &output, const Dictionary &refresher)
{
  // будем удалять уже записанные значения, поэтому сделаем временную копию refresher
  Dictionary rd = refresher;
  string key, value, tail;
  for (;;)
  {
    key.clear();
    value.clear();
    tail.clear();

    switch (read_entry(input, key, value, tail))
    {
    case EntryReadSuccessful:
      // вывести корректную строчку
      {
        // вывести текст ключа
        write(output, key);
        output << " = ";

        // вывести текст значения
        auto p = rd.find(key);
        if (p != rd.end()) // новый из словаря
        {
          write(output, p->second);
          rd.erase (p);
        }
        else // только что прочитанный
        {
          write(output, value);
        }

        // вывести "хвост"
        output << tail << '\n';
      }
      break;

    case EntryReadSkip:
      // вывести пропускаемую строчку
      // если в конце файла идут пустые строки, то они станут разделителем между старой и новой частью
      // -- можно считать сие поведение багом, а можно -- фичей, мне лень с ним бороться...
      output << tail << '\n';
      break;

    case EntryReadEOF:
      // вывести оставшиеся в rd пары ключ-значение и выйти
      if (!rd.empty())
      {
        // пришлось "откусить" последний элемент, чтобы не писать лишний перевод строки
        const auto last = prev(rd.end());
        for (auto p = rd.begin(); p != last; ++p)
        {
          write(output, p->first);
          output << " = ";
          write(output, p->second);
          output << '\n';        
        }

        write(output, last->first);
        output << " = ";
        write(output, last->second);
      }
      return;

    default: {} // ignore
    }
  }
}

}

//
// функции, видимые извне

// загрузить словарь из файла
Dictionary load_dictionary(const char *fn)
{
  ifstream dfile(fn);
  return read_dictionary(dfile);
}


// обновить файл словарём
void refresh_dictionary(const char *fn, const Dictionary &d)
{
  ostringstream output;
  ifstream in_dfile(fn);
  refresh_dictionary(in_dfile, output, d);
  in_dfile.close();

  string data = output.str();
  // удалить все пустые строки в конце
  while (!data.empty() && data.back() == '\n')
    data.pop_back();

  ofstream out_dfile(fn);
  out_dfile << data;
}
